import pytest

from unittest import TestCase
from ztl.core.protocol import Message, Request, State, Task

@pytest.mark.usefixtures("ztl_data")
class TestDataTypes(TestCase):

    def test_message_codec(self):

      encoded = Message.encode(self.msg["scope"], self.msg["state"], self.msg["id"], self.msg["payload"])
      assert type(encoded) is bytes

      decoded = Message.decode(encoded)
      assert type(decoded) is dict

      for field in self.msg.keys():
        assert type(decoded[field]) is str
        assert decoded[field] == self.msg[field]

    def test_task_codec(self):

      encoded = Task.encode(self.task["handler"], self.task["component"], self.task["goal"])
      assert type(encoded) is str

      decoded = Task.decode(encoded)
      assert type(decoded) is dict

      for field in self.task.keys():
        assert type(decoded[field]) is str
        assert decoded[field] == self.task[field]



#!/usr/bin/env python

import sys
import time

import argparse

from ztl.core.server import TaskServer
from ztl.core.protocol import State, Task
from ztl.core.task import ExecutableTask, TaskExecutor, TaskController


class PrintTask(ExecutableTask):

  def __init__(self, request):
    self.active = True
    self.request = request
    self.description = Task.decode(self.request)

  def execute(self):
    print("handler: " + self.description["handler"])
    print("component: " + self.description["component"])
    print("goal: " + self.description["goal"])
    time.sleep(3)
    return "finished"

  def abort(self):
    return False


class TaskTaskController(TaskController):

  def __init__(self):
    self.current_id = 0
    self.running = {}

  def init(self, request):
    self.current_id += 1
    print("Initialising Task ID '%s' (%s)..." % (self.current_id, request))
    self.running[self.current_id] = TaskExecutor(PrintTask, request)
    return self.current_id, "Initiated task '%s' with request: %s" % (self.current_id, request)

  def status(self, mid, request):
    if mid in self.running:
      state = self.running[mid].state()
      print("Status Task ID '%s' (%s): %s." % (mid, request, State.name(state)))
      if state < State.COMPLETED:
        return state, State.name(state)
      else:
        return state, self.running[mid].result()
    else:
      return State.REJECTED, "Invalid ID"

  def abort(self, mid, request):
    if mid in self.running:
      print("Aborting Task ID '%s' (%s)..." % (mid, request))
      success = self.running[mid].stop()
      state = self.running[mid].state()
      return state, success
    else:
      return State.REJECTED, "Invalid ID"

def main_cli():

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("-p", "--port", type=int,
                      help="The port on the local machine that the server should listen to.", required=True)
  parser.add_argument("-s", "--scope", type=str,
                      help="The scope that the server should respond to.", required=True)

  args, unknown = parser.parse_known_args()

  server = TaskServer(args.port)
  server.register(args.scope, TaskTaskController())
  server.listen()


if __name__ == "__main__":

  main_cli()



#!/usr/bin/env python

import sys
import time

from ztl.core.server import TaskServer
from ztl.core.protocol import State
from ztl.core.task import ExecutableTask, TaskExecutor, TaskController


class DummyTask(ExecutableTask):

  def __init__(self, duration):
    self.active = True
    self.duration = duration

  def execute(self):
    start = time.time()
    while self.active and time.time() - start < self.duration:
      time.sleep(.1)
    return self.active

  def abort(self):
    self.active = False
    return True


class SimpleTaskController(TaskController):

  def __init__(self):
    self.current_id = 0
    self.running = {}

  def init(self, request):
    self.current_id += 1
    print("Initialising Task ID '%s' (%s)..." % (self.current_id, request))
    self.running[self.current_id] = TaskExecutor(DummyTask, int(request))
    return self.current_id, "Initiated task '%s' to be active for %s seconds" % (self.current_id, int(request))

  def status(self, mid, request):
    if mid in self.running:
      print("Status Task ID '%s' (%s)..." % (mid, request))
      state = self.running[mid].state()
      if state < State.COMPLETED:
        return state, State.name(state)
      else:
        return state, self.running[mid].result()
    else:
      return State.REJECTED, "Invalid ID"

  def abort(self, mid, request):
    if mid in self.running:
      print("Aborting Task ID '%s' (%s)..." % (mid, request))
      success = self.running[mid].stop()
      state = self.running[mid].state()
      return state, success
    else:
      return State.REJECTED, "Invalid ID"

def main_cli():
  port = sys.argv[1]
  scope = sys.argv[2]
  server = TaskServer(port)
  server.register(scope, SimpleTaskController())
  server.listen()


if __name__ == "__main__":

  main_cli()



#!/usr/bin/env python

import sys

from ztl.core.client import RemoteTask
from ztl.core.protocol import State

def main_cli():
  host = sys.argv[1]
  port = sys.argv[2]
  scope = sys.argv[3]
  request = sys.argv[4]

  print("Connecting to host '%s:%s' at scope '%s'..." % (host, port, scope))
  task = RemoteTask(host, port, scope)
  print("Triggering task with request '%s'..." % request)
  mid, reply = task.trigger(request)

  if mid > 0:
    print("Initialised task with ID '%s'. Reply is '%s'." % (mid, reply))
    print("Waiting maximum 5s for the task to finish...")
    state, reply = task.wait(mid, 5)
    if state < 0:
      print("Could not wait for task with ID '%s'. Reply is '%s'." % (mid, reply))
    elif state <= State.ACCEPTED:
      print("Aborting task with ID '%s'..." % mid)
      state, reply = task.abort(mid)
      if state == State.ABORTED:
        print("Task with ID '%s' aborted. Reply is '%s'." % (mid, reply))
      elif state <= State.ACCEPTED:
        print("Could not abort Task with ID '%s', waiting for completion. Reply is '%s'." % (mid, reply))
        task.wait(mid)
        print("Task with ID '%s' finished after unsuccessful abort signal. Reply is '%s'." % (mid, reply))
    else:
      print("Task with ID '%s' finished while waiting. Result is '%s'." % (mid, reply))
  else:
    print("Task '%s' could not be triggered: '%s'." % reply)

if __name__ == "__main__":

  main_cli()



#!/usr/bin/env python

from distutils.core import setup

setup(name='ztl',
      version='0.0.8',
      description='A thin library relying on zmq to dispatch tasks',
      author='Patrick Holthaus',
      author_email='patrick.holthaus@googlemail.com',
      url='https://gitlab.com/robothouse/rh-user/ztl/',
      package_dir={'':'src'},
      packages=['ztl', 'ztl.core', 'ztl.example', 'ztl.script'],
      scripts=['src/ztl/example/simple_client.py',
            'src/ztl/example/simple_server.py',
            'src/ztl/example/task_client.py',
            'src/ztl/example/task_server.py',
            'src/ztl/script/run_script.py'
      ],
      entry_points={
            'console_scripts': [
                  'ztl_simple_client=ztl.example.simple_client:main_cli',
                  'ztl_simple_server=ztl.example.simple_server:main_cli',
                  'ztl_task_server=ztl.example.task_server:main_cli',
                  'ztl_task_client=ztl.example.task_client:main_cli',
                  'ztl_run_script=ztl.script.run_script:main_cli'
            ]
      },
      install_requires=[
            'oyaml',
            'pyzmq',
            'pytest',
            'pytest-xprocess'
      ]
)



import logging
logging.basicConfig(level=logging.INFO)

from threading import Thread
from ztl.core.protocol import State


class ExecutableTask(object):

  def execute(self):
    pass

  def status(self):
    return None

  def abort(self):
    return False


class TaskController(object):

  def init(self, payload):
    return -1, "Not implemented"


  def status(self, mid, payload):
      return State.REJECTED, "Not implemented"


  def abort(self, mid, payload):
    return State.REJECTED, "Not implemented"


class TaskExecutor(Thread):

  def __init__(self, cls, *parameters):
    Thread.__init__(self)
    self.logger = logging.getLogger(type(cls).__name__)
    self.cls = cls
    self.parameters = parameters
    self.task = None
    self._state = State.INITIATED
    self._result = None
    self._prevent = False
    try:
      self.logger.debug("Initiating task with parameters '%s'...", self.parameters)
      self.task = self.cls(*self.parameters)
      self.start()
    except Exception as e:
      self.logger.debug("Task initialising failed, rejecting task: '%s'", e)
      self._state = State.REJECTED
      self._result = e

  def run(self):
    if not self._prevent:
      self.logger.debug("Accepting and executing task...")
      self._state = State.ACCEPTED
      try:
        self._result = self.task.execute()
        self.logger.debug("Task execution completed successfully.")
        self._state = State.COMPLETED
      except Exception as e:
        self.logger.debug("Task execution failed: '%s'", e)
        self._state = State.FAILED
        self._result = e

    else:
      self._state = State.ABORTED
      logging.warn("Task execution prevented during initialising.")

  def stop(self):
    if self.task is None:
      self.logger.debug("Preventing task from executing...")
      self._prevent = True
      self._state = State.ABORTED
    else:
      self.logger.debug("Aborting task execution...")
      success = self.task.abort()
      if success:
        print("Task aborted successfully.")
        self._state = State.ABORTED
      else:
        self.logger.debug("Task could not be aborted.")
      return success

  def abort(self):
    print("Aborting immediately as requested...")
    self._state = State.ABORTED
    super().abort()

  def state(self):
    if self._state is State.ACCEPTED:
      self._result = self.task.status()
    return self._state

  def result(self):
    return self._result
